import { SqlConnectionPool } from 'sqlite3orm';
import { Sqlite3ModelOptions, Sqlite3PropertyOptions, Sqlite3ForeignKeyOptions } from './sqlite3-options';
export interface SchemaProperty {
    type?: string;
    required?: boolean;
    id?: 0 | 1;
    generated?: 0 | 1;
    sqlite3x?: Sqlite3PropertyOptions;
}
export interface RelationOptions {
    model: string;
    type: string;
    foreignKey: string;
}
export interface Schema {
    name: string;
    options: {
        idInjection: boolean;
        sqlite3x?: Sqlite3ModelOptions;
        foreignKeys?: {
            [key: string]: Sqlite3ForeignKeyOptions;
        };
        relations?: {
            [key: string]: RelationOptions;
        };
    };
    properties: {
        [key: string]: SchemaProperty;
    };
}
export interface Schemas {
    [key: string]: Schema;
}
export interface DiscoverSchemasOptions {
    owner?: string;
    schema?: string;
    schemaName?: string;
    associations?: boolean;
    relations?: boolean;
    visited?: Schemas;
    nameMapper?(type: string, name: string): string;
}
export interface DiscoveredSchema {
    catalog?: string;
    schema: string;
}
export interface DiscoveredTable {
    type: 'table' | 'view';
    name: string;
    owner: string;
}
export declare function defaultNameMapper(type: string, input: string): string;
export declare class DiscoveryService {
    private readonly pool;
    constructor(pool: SqlConnectionPool);
    schemas(): Promise<DiscoveredSchema[]>;
    tables(schemaName: string): Promise<DiscoveredTable[]>;
    table(tableName: string, schemaName?: string, options?: DiscoverSchemasOptions): Promise<Schema>;
    protected propertyType(type: string): string;
}
