"use strict";
// tslint:disable no-implicit-dependencies no-string-literal
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const sqlite3orm_1 = require("sqlite3orm");
const lc_import_1 = require("./lc-import");
const sqlite3_connector_1 = require("./sqlite3-connector");
const utils_1 = require("./utils");
const discovery_service_1 = require("./discovery-service");
/* istanbul ignore next */
function debug(arg, ...args) {
    sqlite3_connector_1.Sqlite3Connector.debug(arg, ...args);
}
/*
 * juggler sql-connector
 */
class Sqlite3JugglerConnector extends lc_import_1.SQLConnector {
    constructor(settings) {
        super(sqlite3_connector_1.SQLITE3_CONNECTOR_NAME);
        this.connector = new sqlite3_connector_1.Sqlite3Connector(settings);
        this.settings = this.connector.settings;
    }
    get pool() {
        return this.connector.pool;
    }
    /* istanbul ignore next */
    static debug(arg, ...args) {
        debug(arg, ...args);
    }
    // *************************************************************************************
    //  connect, disconnect, ping, ...
    // -------------------------------------------------------------------------------------
    /**
     * connect to Sqlite3 database pool
     *
     * @callback [cb] The callback function
     */
    connect(cb) {
        return utils_1.callbackifyOrPromise(this.connector.connect(), cb);
    }
    /**
     * disconnect from Sqlite3 database pool
     *
     * @callback [cb] The callback function
     */
    disconnect(cb) {
        return utils_1.callbackifyOrPromise(this.connector.disconnect(), cb);
    }
    /**
     * ping the database backend
     *
     * @callback [cb] The callback function
     */
    ping(cb) {
        return utils_1.callbackifyOrPromise(this.connector.ping(), cb);
    }
    beginTransaction(isolationLevel, cb) {
        /* istanbul ignore if */
        if (typeof isolationLevel === 'function' && cb === undefined) {
            return this._beginTransaction(isolationLevel);
        }
        else {
            return this._beginTransaction(cb);
        }
    }
    _beginTransaction(cb) {
        return utils_1.callbackifyOrPromise(this.connector.begin(), cb);
    }
    commit(connection, cb) {
        return utils_1.callbackifyOrPromise(this.connector.commit(connection), cb);
    }
    rollback(connection, cb) {
        return utils_1.callbackifyOrPromise(this.connector.rollback(connection), cb);
    }
    executeSQL(sql, params, options, cb) {
        /* istanbul ignore if */
        if (typeof params === 'function' && cb === undefined && options === undefined) {
            cb = params;
            return this._executeSQL(sql, undefined, undefined, cb);
        }
        else {
            /* istanbul ignore if */
            if (typeof options === 'function' && cb === undefined) {
                cb = options;
                return this._executeSQL(sql, params, undefined, cb);
            }
            else {
                return this._executeSQL(sql, params, options, cb);
            }
        }
    }
    _executeSQL(sql, params, options, cb) {
        return utils_1.callbackifyOrPromise(this.promisifiedExecuteSql(sql, params, options), cb);
    }
    promisifiedExecuteSql(sql, params, options) {
        return __awaiter(this, void 0, void 0, function* () {
            params = params || [];
            options = options || {};
            let res;
            try {
                const transaction = options.transaction;
                if (transaction && transaction.connection && transaction.connector === this) {
                    res = yield this.connector.runSQL(transaction.connection, sql, params);
                }
                else {
                    res = yield this.connector.execSql(sql, params);
                }
            }
            catch (err) {
                return Promise.reject(err);
            }
            return Promise.resolve(res);
        });
    }
    getInsertedId(model, result) {
        // the 'result.lastID' is the last inserted RowId
        const metaModel = this.getMetaModel(model);
        /* istanbul ignore else */
        if (result && metaModel.table.mapNameToIdentityField.size === 1) {
            const fld = metaModel.table.mapNameToIdentityField.values().next();
            if (fld.value.dbTypeInfo.typeAffinity === 'INTEGER') {
                // Only if the primary key is integer, it is an alias for the RowId.
                return result.lastID;
            }
        }
        // result.lastID is not the "inserted ID"
        return undefined;
    }
    getCountForAffectedRows(model, result) {
        return (result && result.changes) || 0;
    }
    getPlaceholderForValue(varName) {
        return ':' + varName;
    }
    buildInsertDefaultValues(model, data, options) {
        return 'DEFAULT VALUES';
    }
    fromColumnValue(prop, value) {
        const transform = this.connector.metaModels.getValueTransformer(prop);
        /* istanbul ignore else */
        if (transform) {
            const v = transform.fromDB(value); // using 'undefined' for NULL
            return v !== undefined ? v : this.settings.propertyValueForNULL;
        }
        /* istanbul ignore next */
        throw new Error(`something happened calling fromColumnValue`);
    }
    toColumnValue(prop, value) {
        const transform = this.connector.metaModels.getValueTransformer(prop);
        /* istanbul ignore else */
        if (transform) {
            return transform.toDB(value);
        }
        /* istanbul ignore next */
        return value;
    }
    // *************************************************************************************
    // Mapping:
    // -------------------------------------------------------------------------------------
    /**
     * get the default table/column name for a given model/property name
     * @param name The name
     */
    dbName(modelOrProperty) {
        return this.connector.dbName(modelOrProperty);
    }
    /**
     * escape/quote the table/column name
     * @param name The name
     */
    escapeName(dbName) {
        /* istanbul ignore if */
        if (!dbName) {
            return dbName;
        }
        return '`' + dbName.replace(/\`/g, '``').replace(/\./, '`.`') + '`';
    }
    /**
     * escape a value
     * @param name The name
    escapeValue(value: any): any {
      return value;
    }
     */
    // *************************************************************************************
    // pagination
    // -------------------------------------------------------------------------------------
    static buildLimit(filter) {
        const limit = filter.limit ? (isNaN(filter.limit) ? 0 : filter.limit) : 0;
        const offset = filter.offset ? (isNaN(filter.offset) ? 0 : filter.offset) : 0;
        /* istanbul ignore if */
        if (!limit && !offset) {
            return undefined;
        }
        const res = [];
        /* istanbul ignore else */
        if (limit) {
            res.push(`LIMIT ${limit}`);
        }
        if (offset) {
            res.push(`OFFSET ${offset}`);
        }
        return res.join(' ');
    }
    applyPagination(model, sql, filter) {
        const limitClause = Sqlite3JugglerConnector.buildLimit(filter);
        return limitClause ? sql.merge(limitClause) : sql;
    }
    // *************************************************************************************
    // automigration/autoupdate/isActual
    // -------------------------------------------------------------------------------------
    // TODO(?): add foreign_key_check to automigrate and autoupdate if foreign
    // keys are enabled?
    /**
     * Create the table for the given model
     * @param model The model name
     * @param  [cb] The callback function
     */
    createTable(model, cb) {
        const fn = () => __awaiter(this, void 0, void 0, function* () {
            const metaModel = this.getMetaModel(model, true);
            const conn = yield this.connector.getConnection();
            const factories = [];
            factories.push(() => this.connector.runSQL(conn, metaModel.table.getCreateTableStatement()));
            metaModel.table.mapNameToIDXDef.forEach((idx) => {
                factories.push(() => this.connector.runSQL(conn, metaModel.table.getCreateIndexStatement(idx.name)));
            });
            yield sqlite3orm_1.sequentialize(factories);
            yield conn.close();
        });
        return utils_1.callbackifyOrPromise(fn(), cb);
    }
    /**
     * drop the table for the given model
     * @param model The model name
     * @param  [cb] The callback function
     */
    dropTable(model, cb) {
        const fn = () => __awaiter(this, void 0, void 0, function* () {
            try {
                const metaModel = this.getMetaModel(model);
                yield this.promisifiedExecuteSql(metaModel.table.getDropTableStatement());
            }
            catch (err /* istanbul ignore next */) {
                return Promise.reject(err);
            }
            return;
        });
        return utils_1.callbackifyOrPromise(fn(), cb);
    }
    autoupdate(models, cb) {
        if (!cb && typeof models === 'function') {
            cb = models;
            models = undefined;
        }
        models = models || Object.keys(this._models);
        return this._autoupate(models, cb);
    }
    _autoupate(models, cb) {
        return utils_1.callbackifyOrPromise(this.executeAutoupdate(models), cb);
    }
    executeAutoupdate(model) {
        return __awaiter(this, void 0, void 0, function* () {
            let connection;
            const models = model
                ? typeof model === 'string'
                    ? [model]
                    : model
                : undefined;
            try {
                const tables = [];
                /* istanbul ignore else */
                if (models) {
                    models.forEach((modelName) => {
                        const metaModel = this.getMetaModel(modelName, true);
                        tables.push(metaModel.table);
                    });
                }
                connection = yield this.connector.getConnection();
                const autoUpgrader = new sqlite3orm_1.AutoUpgrader(connection);
                yield autoUpgrader.upgradeTables(tables);
            }
            catch (err) /* istanbul ignore next */ {
                return Promise.reject(err);
            }
            finally {
                if (connection) {
                    yield connection.close();
                }
            }
        });
    }
    isActual(models, cb) {
        if (!cb && typeof models === 'function') {
            cb = models;
            models = undefined;
        }
        models = models || Object.keys(this._models);
        return this._isActual(models, cb);
    }
    _isActual(models, cb) {
        return utils_1.callbackifyOrPromise(this.executeIsActual(models), cb);
    }
    executeIsActual(model) {
        return __awaiter(this, void 0, void 0, function* () {
            let connection;
            const models = model
                ? typeof model === 'string'
                    ? [model]
                    : model
                : undefined;
            try {
                const tables = [];
                /* istanbul ignore else */
                if (models) {
                    models.forEach((modelName) => {
                        const metaModel = this.getMetaModel(modelName, true);
                        tables.push(metaModel.table);
                    });
                }
                connection = yield this.connector.getConnection();
                const autoUpgrader = new sqlite3orm_1.AutoUpgrader(connection);
                const result = yield autoUpgrader.isActual(tables);
                return result;
            }
            catch (err) /* istanbul ignore next */ {
                return Promise.reject(err);
            }
            finally {
                if (connection) {
                    yield connection.close();
                }
            }
        });
    }
    getUpgradeInfo(model) {
        return __awaiter(this, void 0, void 0, function* () {
            let connection;
            try {
                const metaModel = this.getMetaModel(model, true);
                connection = yield this.connector.getConnection();
                const autoUpgrader = new sqlite3orm_1.AutoUpgrader(connection);
                const result = yield autoUpgrader.getUpgradeInfo(metaModel.table);
                return result;
            }
            catch (err) /* istanbul ignore next */ {
                return Promise.reject(err);
            }
            finally {
                if (connection) {
                    yield connection.close();
                }
            }
        });
    }
    discoverDatabaseSchemas(options, cb) {
        if (typeof options === 'function' && !cb) {
            cb = options;
            options = {};
        }
        if (!cb) {
            return;
        }
        const ds = new discovery_service_1.DiscoveryService(this.pool);
        utils_1.callbackify(ds.schemas(), cb);
    }
    discoverModelDefinitions(options, cb) {
        if (typeof options === 'function' && !cb) {
            cb = options;
            options = {};
        }
        if (!cb) {
            return;
        }
        if (!options) {
            options = {};
        }
        const schemaName = options.owner || options.schema || this.settings.schemaName;
        const ds = new discovery_service_1.DiscoveryService(this.pool);
        utils_1.callbackify(ds.tables(schemaName), cb);
    }
    /*
     * discover schema from a given tableName
     * @param tableName
     * @param options
     * @param cb callback
     *
     */
    discoverSchemas(tableName, options, cb) {
        utils_1.callbackify(this.executeDiscoverSchemas(tableName, options), cb);
    }
    /*
     * discover schema from a given tableName
     * @param tableName
     * @param options
     * @returns promise
     *
     */
    executeDiscoverSchemas(tableName, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const ds = new discovery_service_1.DiscoveryService(this.pool);
            if (options.nameMapper === undefined) {
                options.nameMapper = discovery_service_1.defaultNameMapper;
            }
            const schemaName = options.owner || options.schema || this.settings.schemaName;
            const schemaKey = `${schemaName}.${tableName}`;
            options.visited = options.visited || {};
            if (options.visited[schemaKey]) {
                return options.visited;
            }
            try {
                const schema = yield ds.table(tableName, schemaName, options);
                if (options.visited[schemaKey]) {
                    return options.visited;
                }
                options.visited[schemaKey] = schema;
                debug(`disovered: ${schemaKey}`);
                const fks = schema.options.foreignKeys;
                if (!fks || Object.keys(fks).length === 0 || (!options.associations && !options.relations)) {
                    return options.visited;
                }
                schema.options.relations = {};
                Object.keys(fks).forEach((fkName) => {
                    const fk = fks[fkName];
                    // NOTE: fkName is a generic name not intended for use in model definitions
                    // => defering fkModelName from refTable
                    const fkModelName = options.nameMapper
                        ? options.nameMapper('fk', fk.refTable)
                        : fk.refTable;
                    let fkKey = fkModelName;
                    let fkLastId = 0;
                    // tslint:disable-next-line no-non-null-assertion
                    while (schema.options.relations[fkKey]) {
                        fkKey = `${fkModelName}${++fkLastId}`;
                    }
                    schema.options.relations[fkKey] = {
                        model: options.nameMapper ? options.nameMapper('table', fk.refTable) : fk.refTable,
                        type: 'belongsTo',
                        foreignKey: fk.properties,
                    };
                });
                // NOTE: because of sqlite3 limitation, all referenced tables are in the same schema as the referencing table
                yield sqlite3orm_1.sequentialize(Object.keys(fks).map((fkName) => () => this.executeDiscoverSchemas(fks[fkName].refTable, options)));
                return options.visited;
            }
            catch (err) {
                return Promise.reject(err);
            }
        });
    }
    // *************************************************************************************
    // model definitions
    // -------------------------------------------------------------------------------------
    modelNames() {
        const lbModelDef = this._models;
        return lbModelDef ? Object.keys(lbModelDef) : [];
    }
    getMetaModel(modelName, recreate) {
        const lbModelDef = this._models[modelName];
        /* istanbul ignore if */
        if (!lbModelDef) {
            throw new Error(`model '${modelName}' not found`);
        }
        return this.connector.getMetaModelFromJuggler(modelName, lbModelDef, recreate);
    }
    destroyMetaModel(modelName) {
        this.connector.destroyMetaModel(modelName);
    }
    destroyAllMetaModels() {
        this.connector.destroyAllMetaModels();
    }
}
exports.Sqlite3JugglerConnector = Sqlite3JugglerConnector;
/**
 *
 * Initialize the Sqlite3Connector for the given data source
 *
 * @param dataSource The dataSource
 * @callback [cb] The callback function
 */
function initialize(dataSource, cb) {
    /* istanbul ignore if */
    if (dataSource.settings && !!dataSource.settings.debug) {
        sqlite3_connector_1.Sqlite3Connector.debugEnabled = true;
    }
    debug(`initialize ${sqlite3_connector_1.SQLITE3_CONNECTOR_NAME} juggler connector ...`);
    const connector = new Sqlite3JugglerConnector(dataSource.settings);
    dataSource.connector = connector;
    dataSource.connector.dataSource = dataSource;
    /* istanbul ignore else */
    if (cb) {
        if (dataSource.settings && dataSource.settings.lazyConnect) {
            cb();
        }
        else {
            debug(`initial connect`);
            connector.connect((err) => {
                /* istanbul ignore if */
                if (err) {
                    debug(`initial connect failed: `, err);
                }
                else {
                    debug(`initial connect succeeded`);
                }
                debug(`initialized ${sqlite3_connector_1.SQLITE3_CONNECTOR_NAME} juggler connector`);
                cb(err);
            });
            return;
        }
    }
    debug(`initialized ${sqlite3_connector_1.SQLITE3_CONNECTOR_NAME} juggler connector`);
}
exports.initialize = initialize;
//# sourceMappingURL=sqlite3-juggler-connector.js.map